use cosmrs::{
    proto::prost::Message,
    rpc::Client,
    tx::{self, BodyBuilder, Fee, Msg, SignDoc, SignerInfo},
    AccountId, Any,
};
use icepick_module::Module;
use serde::{Deserialize, Serialize};
use std::str::FromStr;

use cosmrs::crypto::secp256k1;

mod coin_denoms;
mod remote_serde;

#[derive(thiserror::Error, Debug)]
pub enum Error {}

#[derive(Serialize, Deserialize, Debug)]
pub struct GetChainInfo {
    chain_name: String,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct GenerateWallet {
    account: Option<String>,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct GetWalletAddress {
    blockchain_config: coin_denoms::Blockchain,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct GetAccountData {
    account_id: String,

    blockchain_config: coin_denoms::Blockchain,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct AwaitFunds {
    address: String,
    denom_name: String,
    amount: String,

    blockchain_config: coin_denoms::Blockchain,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Transfer {
    amount: String,
    denom: String,
    to_address: String,
    from_account: Option<String>,
    from_address: String,
    // TODO: find a way to simulate transaction and calculate gas necessary
    // for now, 0.01KYVE seems to be a reasonable mainnet number?
    // for testing purposes, i'm gonna go much lower. 0.0001.
    gas_factor: Option<String>,

    blockchain_config: coin_denoms::Blockchain,
}

#[derive(Deserialize, Serialize, Debug)]
pub struct Sign {
    fee: remote_serde::Fee,
    tx_messages: Vec<Any>,
    account_number: u64,
    sequence_number: u64,

    blockchain_config: coin_denoms::Blockchain,
}

#[derive(Deserialize, Serialize, Debug)]
pub struct Broadcast {
    transaction: Vec<u8>,

    blockchain_config: coin_denoms::Blockchain,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Request {
    derived_keys: Option<Vec<[u8; 32]>>,

    #[serde(flatten)]
    operation: Operation,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(tag = "operation", content = "values", rename_all = "kebab-case")]
#[allow(clippy::large_enum_variant)]
pub enum Operation {
    GetChainInfo(GetChainInfo),
    GenerateWallet(GenerateWallet),
    GetWalletAddress(GetWalletAddress),
    GetAccountData(GetAccountData),
    AwaitFunds(AwaitFunds),
    Transfer(Transfer),
    Sign(Sign),
    Broadcast(Broadcast),
}

pub fn run_async<F: std::future::Future>(f: F) -> F::Output {
    tokio::runtime::Builder::new_multi_thread()
        .worker_threads(2)
        .enable_all()
        .build()
        .unwrap()
        .block_on(f)
}

fn decode<T: cosmrs::proto::prost::Message + cosmrs::proto::prost::Name + std::default::Default>(
    _type: &str,
    value: &[u8],
) -> Result<T, String> {
    // move past the first `/`.
    let _type = &_type[1..];
    let full_name = T::full_name();
    if _type != full_name {
        return Err(format!(
            "given type {_type} does not match expected type {full_name}"
        ));
    }

    T::decode(value).map_err(|e| e.to_string())
}

async fn abci_query<
    Response: cosmrs::proto::prost::Message + cosmrs::proto::prost::Name + std::default::Default,
>(
    client: &(impl Client + Sync),
    path: &'static str,
    request_data: Option<&impl cosmrs::proto::prost::Message>,
    height: Option<cosmrs::tendermint::block::Height>,
    prove: bool,
) -> Result<Response, Box<dyn std::error::Error>> {
    let data = request_data.map(Message::encode_to_vec).unwrap_or_default();
    let response = client
        .abci_query(Some(path.to_string()), data, height, prove)
        .await?;
    let parsed = Response::decode(&*response.value)?;
    Ok(parsed)
}

pub struct Cosmos;

impl Module for Cosmos {
    type Error = Error;
    type Request = Request;

    fn describe_operations() -> Vec<icepick_module::help::Operation> {
        use icepick_module::help::*;
        let account = Argument::builder()
            .name("account")
            .description("The derivation index for the account.")
            .r#type(ArgumentType::Optional)
            .build();

        let address_prefix = Argument::builder()
            .name("address_prefix")
            .description("Prefix for the wallet (`cosmos`, `kyve`, etc.).")
            .r#type(ArgumentType::Required)
            .build();

        let get_chain_info = Operation::builder()
            .name("get-chain-info")
            .description("Get information for a given chain")
            .build()
            .argument(
                &Argument::builder()
                    .name("chain_name")
                    .description("Name of the blockchain")
                    .r#type(ArgumentType::Required)
                    .build(),
            );

        let generate_wallet = Operation::builder()
            .name("generate-wallet")
            .description("Generate a wallet for the given account.")
            .build()
            .argument(&account);

        let get_wallet_address = Operation::builder()
            .name("get-wallet-address")
            .description("Get the address for a given wallet.")
            .build()
            .argument(&address_prefix);

        let get_account_info = Operation::builder()
            .name("get-account-data")
            .description("Get the account number and sequence number for an account.")
            .build()
            .argument(
                &Argument::builder()
                    .name("account_id")
                    .description("The account ID to get account information for.")
                    .r#type(ArgumentType::Required)
                    .build(),
            );

        let await_funds = icepick_module::help::Operation {
            name: "await-funds".to_string(),
            description: "Await a minimum amount of funds in an account".to_string(),
            arguments: vec![
                Argument {
                    name: "address".to_string(),
                    description: "The address to monitor".to_string(),
                    r#type: ArgumentType::Required,
                },
                Argument {
                    name: "denom_name".to_string(),
                    description: "The denomination of coin to monitor".to_string(),
                    r#type: ArgumentType::Required,
                },
                Argument {
                    name: "amount".to_string(),
                    description: "The amount of the minimum denomination to await".to_string(),
                    r#type: ArgumentType::Required,
                },
            ],
        };

        let transfer = Operation::builder()
            .name("transfer")
            .description("Transfer coins from one address to another.")
            .build()
            .argument(
                &Argument::builder()
                    .name("from_address")
                    .description("The address to send coins from.")
                    .r#type(ArgumentType::Required)
                    .build(),
            )
            .argument(
                &Argument::builder()
                    .name("to_address")
                    .description("The address to send coins to.")
                    .r#type(ArgumentType::Required)
                    .build(),
            )
            .argument(
                &Argument::builder()
                    .name("amount")
                    .description("The amount of coins to send.")
                    .r#type(ArgumentType::Required)
                    .build(),
            )
            .argument(
                &Argument::builder()
                    .name("denom")
                    .description("The denomination of coin to send.")
                    .r#type(ArgumentType::Required)
                    .build(),
            )
            .argument(
                &Argument::builder()
                    .name("gas_factor")
                    .description("The factor to multiply the default gas amount by.")
                    .r#type(ArgumentType::Optional)
                    .build(),
            );

        let sign = Operation::builder()
            .name("sign")
            .description("Sign a previously-generated transaction.")
            .build()
            .argument(
                &Argument::builder()
                    .name("account_number")
                    .description("The sequence number for the given public key")
                    .r#type(ArgumentType::Required)
                    .build(),
            )
            .argument(
                &Argument::builder()
                    .name("sequence_number")
                    .description("The account number for the given public key")
                    .r#type(ArgumentType::Required)
                    .build(),
            );

        let broadcast = Operation::builder()
            .name("broadcast")
            .description("Broadcast a transaction.")
            .build();

        vec![
            get_chain_info,
            generate_wallet,
            get_wallet_address,
            get_account_info,
            await_funds,
            transfer,
            sign,
            broadcast,
        ]
    }

    fn handle_request(request: Self::Request) -> Result<serde_json::Value, Self::Error> {
        let Request {
            operation,
            derived_keys: _,
        } = request;

        match operation {
            Operation::GetChainInfo(GetChainInfo { chain_name }) => {
                let chains = coin_denoms::default_chains();

                let chain = chains
                    .iter()
                    .find(|chain| chain.chain_name == chain_name || chain.chain_id == chain_name);

                Ok(serde_json::json!({
                    "blob": {
                        "blockchain_config": chain,
                    },
                }))
            }
            Operation::GenerateWallet(GenerateWallet { account }) => {
                let account = u32::from_str(account.as_deref().unwrap_or("0")).unwrap();
                Ok(serde_json::json!({
                    "blob": {},
                    "derivation_accounts": [(account | 1 << 31)],
                }))
            }
            Operation::GetWalletAddress(GetWalletAddress { blockchain_config }) => {
                // NOTE: panics if doesn't exist
                let key = request.derived_keys.unwrap()[0];
                let privkey = secp256k1::SigningKey::from_slice(&key).unwrap();
                let pubkey = privkey.public_key();
                let sender_account_id = pubkey
                    .account_id(&blockchain_config.bech32_config.account_address_prefix)
                    .unwrap();
                Ok(serde_json::json!({
                    "blob": {
                        "pubkey": sender_account_id,
                    }
                }))
            }
            Operation::GetAccountData(GetAccountData {
                account_id,
                blockchain_config,
            }) => {
                use cosmrs::proto::cosmos::auth::v1beta1::*;

                let response = run_async(async {
                    let client =
                        cosmrs::rpc::HttpClient::new(blockchain_config.rpc_url.as_str()).unwrap();
                    let query = QueryAccountRequest {
                        address: account_id,
                    };
                    let response: QueryAccountResponse = abci_query(
                        &client,
                        "/cosmos.auth.v1beta1.Query/Account",
                        Some(&query),
                        None,
                        false,
                    )
                    .await
                    .unwrap();
                    response
                });
                let Any { type_url, value } = response.account.unwrap();
                let account: BaseAccount = decode(&type_url, &value).unwrap();
                Ok(serde_json::json!({
                    "blob": {
                        "account_number": account.account_number,
                        "sequence_number": account.sequence,
                    }
                }))
            }
            Operation::AwaitFunds(AwaitFunds {
                address,
                denom_name,
                amount,
                blockchain_config,
            }) => {
                use cosmrs::proto::cosmos::bank::v1beta1::*;

                // Check if given denom is min denom or normal and adjust accordingly
                let Some(relevant_denom) = blockchain_config.currencies.iter().find(|c| {
                    [&c.coin_denom, &c.coin_minimal_denom]
                        .iter()
                        .any(|name| **name == denom_name)
                }) else {
                    panic!("{denom_name} not in {blockchain_config:?}");
                };

                let amount = f64::from_str(&amount).unwrap();
                let adjusted_amount = if relevant_denom.coin_denom == denom_name {
                    amount * 10f64.powi(i32::from(relevant_denom.coin_decimals))
                } else if relevant_denom.coin_minimal_denom == denom_name {
                    amount
                } else {
                    unreachable!("broke invariant: check denom checker");
                } as u128;

                let coin = run_async(async {
                    let client =
                        cosmrs::rpc::HttpClient::new(blockchain_config.rpc_url.as_str()).unwrap();

                    loop {
                        let response: QueryBalanceResponse = abci_query(
                            &client,
                            "/cosmos.bank.v1beta1.Query/Balance",
                            Some(&QueryBalanceRequest {
                                address: address.clone(),
                                denom: relevant_denom.coin_minimal_denom.clone(),
                            }),
                            None,
                            false,
                        )
                        .await
                        .unwrap();

                        if let Some(coin) = response.balance {
                            let amount = u128::from_str(&coin.amount).unwrap();
                            if amount >= adjusted_amount {
                                break coin;
                            }
                        }
                        tokio::time::sleep(std::time::Duration::from_secs(10)).await;
                    }
                });
                let cosmrs::proto::cosmos::base::v1beta1::Coin { denom, amount } = coin;
                Ok(serde_json::json!({
                    "blob": {
                        "balance": {
                            "denom": denom,
                            "amount": u128::from_str(&amount).unwrap(),
                        },
                    }
                }))
            }
            Operation::Transfer(Transfer {
                amount,
                denom,
                to_address,
                from_account: _,
                from_address,
                gas_factor,
                blockchain_config,
            }) => {
                // Check if given denom is min denom or normal and adjust accordingly
                let Some(relevant_denom) = blockchain_config.currencies.iter().find(|c| {
                    [&c.coin_denom, &c.coin_minimal_denom]
                        .iter()
                        .any(|name| **name == denom)
                }) else {
                    panic!("{denom} not in {blockchain_config:?}");
                };

                let gas_factor = gas_factor
                    .as_deref()
                    .map(f64::from_str)
                    .transpose()
                    .unwrap()
                    .unwrap_or(1.0);

                let amount = f64::from_str(&amount).unwrap();
                let adjusted_amount = if relevant_denom.coin_denom == denom {
                    amount * 10f64.powi(i32::from(relevant_denom.coin_decimals))
                } else if relevant_denom.coin_minimal_denom == denom {
                    amount
                } else {
                    unreachable!("broke invariant: check denom checker");
                } as u128;

                let from_id = AccountId::from_str(&from_address).unwrap();
                let to_id = AccountId::from_str(&to_address).unwrap();

                let coin = cosmrs::Coin {
                    denom: relevant_denom.coin_minimal_denom.parse().unwrap(),
                    amount: adjusted_amount,
                };

                let msg_send = cosmrs::bank::MsgSend {
                    from_address: from_id,
                    to_address: to_id,
                    amount: vec![coin.clone()],
                }
                .to_any()
                .unwrap();

                let expected_gas = 100_000u64;
                // convert gas "price" to minimum denom,
                // multiply by amount of gas required,
                // multiply by gas factor if necessary.
                let expected_fee = blockchain_config.gas_price_step.high
                    // * dbg!(10f64.powi(relevant_denom.coin_decimals as i32))
                    * expected_gas as f64
                    * gas_factor;

                let fee_coin = cosmrs::Coin {
                    denom: relevant_denom.coin_minimal_denom.parse().unwrap(),
                    amount: expected_fee as u128,
                };

                let fee = Fee::from_amount_and_gas(fee_coin, expected_gas);

                #[allow(clippy::identity_op)]
                Ok(serde_json::json!({
                    "blob": {
                        "fee": remote_serde::Fee::from(&fee),
                        // TODO: Body does not implement Serialize and
                        // needs to be constructed in Sign
                        "tx_messages": [msg_send],
                        // re-export, but in general this should be copied over
                        // using workflows
                        "blockchain_config": blockchain_config,
                    },
                    "derivation_accounts": [0u32 | 1 << 31],
                }))
            }
            Operation::Sign(Sign {
                fee,
                tx_messages,
                account_number,
                sequence_number,
                blockchain_config,
            }) => {
                let key = request.derived_keys.unwrap()[0];
                let privkey = secp256k1::SigningKey::from_slice(&key).unwrap();

                let fee = cosmrs::tx::Fee::from(&fee);
                let tx_body = BodyBuilder::new().msgs(tx_messages).finish();

                let auth_info =
                    SignerInfo::single_direct(Some(privkey.public_key()), sequence_number)
                        .auth_info(fee);
                let sign_doc = SignDoc::new(
                    &tx_body,
                    &auth_info,
                    &blockchain_config.chain_id.parse().unwrap(),
                    account_number,
                )
                .unwrap();
                let signed_tx = sign_doc.sign(&privkey).unwrap();
                Ok(serde_json::json!({
                    "blob": {
                        "transaction": signed_tx.to_bytes().unwrap(),
                        "blockchain_config": blockchain_config,
                    }
                }))
            }
            Operation::Broadcast(Broadcast {
                transaction,
                blockchain_config,
            }) => {
                let tx = tx::Raw::from_bytes(&transaction).unwrap();

                let response = run_async(async {
                    let client =
                        cosmrs::rpc::HttpClient::new(blockchain_config.rpc_url.as_str()).unwrap();

                    // broadcast_tx_sync awaits CheckTx, so we know that, at the bare minimum, the
                    // transaction is valid.
                    //
                    // TODO: make this expect() into a keyfork_bug!(). An error in this area of
                    // code may represent a bug in either our code or the blockchain's code.
                    // We _should_ get an Err code if the error was with the transaction data, such
                    // as us being out of gas, or the transaction being malformed.
                    client
                        .broadcast_tx_sync(tx.to_bytes().unwrap())
                        .await
                        .expect("The server encountered a fatal error while processing the request")
                });
                match response.code {
                    cosmrs::tendermint::abci::Code::Ok => {
                        // attempt to get transaction URL / blockchain explorer URL
                        Ok(serde_json::json!({
                            "blob": {
                                "status": "send_and_confirm",
                                "success": response.hash.to_string(),
                                "url": blockchain_config.explorer_url_format.replace("%s", response.hash.to_string().as_str()),
                            }
                        }))
                    }
                    cosmrs::tendermint::abci::Code::Err(non_zero) => Ok(serde_json::json!({
                        "blob": {
                            "status": "send_and_confirm",
                            "error": response.log,
                            "error_code": non_zero.get(),
                        }
                    })),
                }
            }
        }
    }
}
