use icepick_workflow::{InvocableOperation, OperationResult, Workflow};
use keyfork_derive_util::{request::DerivationAlgorithm, DerivationPath};
use serde_json::Value;
use std::{
    collections::HashMap,
    io::Write,
    process::{Command, Stdio},
};

use super::{derive_keys, get_command, Commands, ModuleConfig, Operation};

pub type StringMap = std::collections::HashMap<String, String>;

#[derive(Clone, Debug)]
struct CLIOperation {
    /// The name of the operation (i.e. `transfer-token`).
    name: String,

    /// The binary to invoke the operation.
    binary: String,

    /// Information describing the invocation requirements of the operation.
    operation: Operation,

    /// The derivation algorithm used when deriving keys for the operation.
    derivation_algorithm: Option<DerivationAlgorithm>,

    /// The derivation prefix used when deriving keys for the operation.
    derivation_prefix: Option<DerivationPath>,
}

impl InvocableOperation for CLIOperation {
    fn invoke(&self, input: &HashMap<String, Value>, derived_keys: &[Vec<u8>]) -> OperationResult {
        let (command, args) = get_command(&self.binary);

        let json = serde_json::json!({
            "operation": self.operation.name,
            "values": input,
            "derived_keys": derived_keys,
        });

        let mut child = Command::new(command)
            .args(args)
            .stdin(Stdio::piped())
            .stdout(Stdio::piped())
            .spawn()
            .unwrap();

        let mut child_input = child.stdin.take().unwrap();
        serde_json::to_writer(&mut child_input, &json).unwrap();
        child_input
            .write_all(b"\n{\"operation\": \"exit\"}\n")
            .unwrap();

        let result = child.wait_with_output().unwrap();
        if !result.status.success() {
            panic!(
                "Bad exit ({}: {}): {}",
                &self.binary,
                &self.operation.name,
                String::from_utf8_lossy(&result.stderr)
            );
        }

        let output = result.stdout;
        let json: OperationResult = serde_json::from_slice(&output).expect("valid json");
        json
    }

    fn name(&self) -> &String {
        &self.name
    }

    fn argument_names(&self) -> impl Iterator<Item = &String> {
        self.operation.arguments.iter().map(|i| &i.name)
    }

    fn derivation_configuration(&self) -> Option<(&DerivationAlgorithm, &DerivationPath)> {
        self.derivation_algorithm
            .as_ref()
            .zip(self.derivation_prefix.as_ref())
    }
}

pub fn generate_command(workflow: &Workflow) -> clap::Command {
    let mut command = clap::Command::new(&workflow.name).arg(clap::arg!(
            --"input-file" [FILE]
            "A file containing any inputs not passed on the command line"
    ));
    for input in &workflow.inputs {
        // can also be included in the JSON file, so we won't mark this as required.
        let arg = clap::Arg::new(input)
            .required(false)
            .long(input.replace('_', "-"))
            .value_name(input.to_uppercase());
        command = command.arg(arg);
    }
    for input in &workflow.optional_inputs {
        let arg = clap::Arg::new(input)
            .required(false)
            .long(input.replace('_', "-"))
            .value_name(input.to_uppercase());
        command = command.arg(arg);
    }
    command
}

fn load_inputs<T: AsRef<str> + Into<String> + std::fmt::Display>(
    inputs: impl IntoIterator<Item = T>,
    optional_inputs: impl IntoIterator<Item = T>,
    matches: &clap::ArgMatches,
) -> StringMap {
    let mut map = StringMap::default();
    let input_file: Option<StringMap> = matches
        .get_one::<std::path::PathBuf>("input-file")
        .and_then(|p| std::fs::File::open(p).ok())
        .and_then(|f| serde_json::from_reader(f).ok());
    for input in inputs {
        match matches.get_one::<String>(input.as_ref()) {
            Some(value) => {
                map.insert(input.into(), value.clone());
                continue;
            }
            None => {
                if let Some(value) = input_file.as_ref().and_then(|f| f.get(input.as_ref())) {
                    map.insert(input.into(), value.clone());
                    continue;
                }
            }
        }
        panic!("Required workflow input was not found: {input}");
    }

    for input in optional_inputs {
        match matches.get_one::<String>(input.as_ref()) {
            Some(value) => {
                map.insert(input.into(), value.clone());
                continue;
            }
            None => {
                if let Some(value) = input_file.as_ref().and_then(|f| f.get(input.as_ref())) {
                    map.insert(input.into(), value.clone());
                    continue;
                }
            }
        }
    }

    map
}

fn load_operations(commands: Commands, config: &[ModuleConfig]) -> Vec<CLIOperation> {
    let mut operations = vec![];

    for (module_name, module_binary, module_operations) in commands {
        for operation in module_operations {
            let operation_name = &operation.name;
            let module_config = config.iter().find(|conf| conf.name == *module_name);
            let io = CLIOperation {
                name: format!("{module_name}-{operation_name}"),
                binary: module_binary.clone(),
                operation: operation.clone(),
                derivation_algorithm: module_config.and_then(|m| m.algorithm.clone()),
                derivation_prefix: module_config.and_then(|m| m.derivation_prefix.clone()),
            };
            operations.push(io);
        }
    }

    operations
}

pub fn handle(
    workflow: &Workflow,
    matches: &clap::ArgMatches,
    modules: Commands,
    config: &[ModuleConfig],
) {
    let inputs = load_inputs(&workflow.inputs, &workflow.optional_inputs, matches);
    let data: HashMap<String, Value> = inputs
        .into_iter()
        .map(|(k, v)| (k, Value::String(v)))
        .collect();

    let operations = load_operations(modules, config);

    if matches.get_flag("simulate-workflow") {
        let reports = workflow
            .simulate_workflow(data.into_keys().collect(), &operations)
            .expect("Simulation failure");
        for report in reports {
            println!("{report}");
        }
        return;
    }

    let result = workflow
        .run_workflow(data, &operations, &derive_keys)
        .expect("Invocation failure");
    println!("{}", serde_json::to_string(&result).expect("valid JSON"));
}
